# -*- coding: utf-8 -*-
"""
Class for transmitting content metadata to Cornerstone.
"""
from itertools import islice

from django.conf import settings

from integrated_channels.cornerstone.client import CornerstoneAPIClient
from integrated_channels.integrated_channel.transmitters.content_metadata import ContentMetadataTransmitter
from integrated_channels.utils import chunks


class CornerstoneContentMetadataTransmitter(ContentMetadataTransmitter):
    """
    This transmitter prepares content metadata to be consumed by Cornerstone.
    """

    def __init__(self, enterprise_configuration, client=CornerstoneAPIClient):
        super().__init__(
            enterprise_configuration=enterprise_configuration,
            client=client
        )

    def transmit(self, payload, **kwargs):
        """
        Prepare content metadata items for cornerstone API consumer. Generate, update or delete a
        ContentMetadataItemTransmission object depending on the state of the consumers catalog metadata.

        Args:
            payload (OrderedDict): course metadata dictionary corresponding to the result of the content metadata
            exporter's export.
        """
        items_to_create, items_to_update, items_to_delete, transmission_map, items_catalog_updated_times = \
            self._partition_items(payload)
        self._transmit_delete(items_to_delete)
        self._transmit_create(items_to_create, items_catalog_updated_times)
        self._transmit_update(items_to_update, transmission_map, items_catalog_updated_times)
        return self._prepare_items_for_transmission(payload)

    def _transmit_create(self, channel_metadata_item_map, items_catalog_updated_times):
        """
        Generate the appropriate ContentMetadataItemTransmission objects according to the channel_metadata_item_map

        Args:
            channel_metadata_item_map (dict): A dictionary representation of the courses to be created based on the
            exported content metadata.

            items_catalog_updated_times (dict): Mapping between course keys and the last updated time of the associated
                enterprise catalog
        """
        transmission_limit = settings.INTEGRATED_CHANNELS_API_CHUNK_TRANSMISSION_LIMIT.get(
            self.enterprise_configuration.channel_code()
        )
        create_chunk_items = chunks(channel_metadata_item_map, self.enterprise_configuration.transmission_chunk_size)
        for chunk in islice(create_chunk_items, transmission_limit):
            self._create_transmissions(chunk, items_catalog_updated_times)

    def _transmit_update(self, channel_metadata_item_map, transmission_map, items_catalog_updated_times):
        """
        Update the appropriate ContentMetadataItemTransmission objects according to the channel_metadata_item_map and
        transmission_map

        Args:
            channel_metadata_item_map (dict): A dictionary representation of the courses to be updated based on the
            exported content metadata.

            transmission_map (dict): A dictionary mapping of which transmission items to be updated.

            items_catalog_updated_times (dict): Mapping between course keys and the last updated time of the associated
                enterprise catalog
            """
        transmission_limit = settings.INTEGRATED_CHANNELS_API_CHUNK_TRANSMISSION_LIMIT.get(
            self.enterprise_configuration.channel_code()
        )
        update_chunk_items = chunks(channel_metadata_item_map, self.enterprise_configuration.transmission_chunk_size)
        for chunk in islice(update_chunk_items, transmission_limit):
            self._update_transmissions(chunk, transmission_map, items_catalog_updated_times)

    def _transmit_delete(self, channel_metadata_item_map):
        """
        Remove the appropriate ContentMetadataItemTransmission objects according to the channel_metadata_item_map

        Args:
            channel_metadata_item_map (dict): A dictionary representation of the courses to be removed based on the
            exported content metadata.
        """
        transmission_limit = settings.INTEGRATED_CHANNELS_API_CHUNK_TRANSMISSION_LIMIT.get(
            self.enterprise_configuration.channel_code()
        )
        delete_chunk_items = chunks(channel_metadata_item_map, self.enterprise_configuration.transmission_chunk_size)
        for chunk in islice(delete_chunk_items, transmission_limit):
            self._delete_transmissions(chunk.keys())

    def _prepare_items_for_transmission(self, channel_metadata_items):
        """
        Format the content metadata to what CSOD consumers expect.
        """
        course_list = [
            item.channel_metadata
            for item in channel_metadata_items.values()
        ]
        return course_list



# -*- coding: utf-8 -*-
"""
Class for transmitting content metadata to SuccessFactors.
"""

import logging
from itertools import islice

from django.conf import settings

from integrated_channels.exceptions import ClientError
from integrated_channels.integrated_channel.transmitters.content_metadata import ContentMetadataTransmitter
from integrated_channels.sap_success_factors.client import SAPSuccessFactorsAPIClient
from integrated_channels.utils import chunks

LOGGER = logging.getLogger(__name__)


class SapSuccessFactorsContentMetadataTransmitter(ContentMetadataTransmitter):
    """
    This transmitter transmits exported content metadata to SAPSF.
    """

    def __init__(self, enterprise_configuration, client=SAPSuccessFactorsAPIClient):
        """
        Use the ``SAPSuccessFactorsAPIClient`` for content metadata transmission to SAPSF.
        """
        super().__init__(
            enterprise_configuration=enterprise_configuration,
            client=client
        )

    def transmit(self, payload, **kwargs):
        """
        Transmit content metadata items to the integrated channel.
        """
        items_to_create, items_to_update, items_to_delete, transmission_map, items_catalog_updated_times = \
            self._partition_items(payload)
        self._prepare_items_for_delete(items_to_delete)
        prepared_items = {}
        prepared_items.update(items_to_create)
        prepared_items.update(items_to_update)
        prepared_items.update(items_to_delete)

        chunk_items = chunks(prepared_items, self.enterprise_configuration.transmission_chunk_size)
        transmission_limit = settings.INTEGRATED_CHANNELS_API_CHUNK_TRANSMISSION_LIMIT.get(
            self.enterprise_configuration.channel_code()
        )
        for chunk in islice(chunk_items, transmission_limit):
            chunked_items = list(chunk.values())
            try:
                self.client.update_content_metadata(self._serialize_items(chunked_items))
            except ClientError as exc:
                LOGGER.error(
                    'Failed to update [%s] content metadata items for integrated channel [%s] [%s]',
                    len(chunked_items),
                    self.enterprise_configuration.enterprise_customer.name,
                    self.enterprise_configuration.channel_code(),
                )
                LOGGER.exception(exc)

                # Remove the failed items from the create/update/delete dictionaries,
                # so ContentMetadataItemTransmission objects are not synchronized for
                # these items below.
                self._remove_failed_items(chunked_items, items_to_create, items_to_update, items_to_delete)

        # If API transmission limit is set then mark the rest of the items as not transmitted.
        # Since, chunk_items is a generator and we have already iterated through the items that need to
        # be transmitted. Rest of the items are the ones that need to marked as not transmitted.
        for chunk in chunk_items:
            chunked_items = list(chunk.values())
            self._remove_failed_items(chunked_items, items_to_create, items_to_update, items_to_delete)

        self._create_transmissions(items_to_create, items_catalog_updated_times)
        self._update_transmissions(items_to_update, transmission_map, items_catalog_updated_times)
        self._delete_transmissions(list(items_to_delete.keys()))

    def _remove_failed_items(self, failed_items, items_to_create, items_to_update, items_to_delete):
        """
        Remove content metadata items from the `items_to_create`, `items_to_update`, `items_to_delete` dicts.

        Arguments:
            failed_items (list): Failed Items to be removed.
            items_to_create (dict): dict containing the items created successfully.
            items_to_update (dict): dict containing the items updated successfully.
            items_to_delete (dict): dict containing the items deleted successfully.
        """
        for item in failed_items:
            content_metadata_id = item['courseID']
            items_to_create.pop(content_metadata_id, None)
            items_to_update.pop(content_metadata_id, None)
            items_to_delete.pop(content_metadata_id, None)

    def _prepare_items_for_transmission(self, channel_metadata_items):
        return {
            'ocnCourses': channel_metadata_items
        }

    def _prepare_items_for_delete(self, items_to_delete):
        """
        Set status to INACTIVE for items that should be deleted.
        """
        for item in items_to_delete.values():
            item['status'] = 'INACTIVE'



# -*- coding: utf-8 -*-
"""
Assist integrated channels with retrieving content metadata.

Module contains resources for integrated channels to retrieve all the
metadata for content contained in the catalogs associated with a particular
enterprise customer.
"""

import json
from collections import OrderedDict
from datetime import datetime
from logging import getLogger

import pytz

from django.apps import apps
from django.utils import dateparse

from enterprise.api_client.enterprise_catalog import EnterpriseCatalogApiClient
from enterprise.utils import get_content_metadata_item_id
from integrated_channels.integrated_channel.exporters import Exporter

LOGGER = getLogger(__name__)


class ContentMetadataExporter(Exporter):
    """
    Base class for content metadata exporters.
    """

    # DATA_TRANSFORM_MAPPING is used to map the content metadata field names expected by the integrated channel
    # to the edX content metadata schema. The values contained in the dict will be used as keys to access values
    # in each content metadata item dict which is being exported.
    #
    # Example:
    #     {
    #         'contentID': 'key',
    #         'courseTitle': 'title'
    #     }
    #
    #     Defines a transformation of the content metadata item to:
    #
    #     {
    #         'contentID': content_metadata_item['key'],
    #         'courseTitle': content_metadata_item['title']
    #     }
    #
    # Subclasses should override this class variable. By default, the edX content metadata schema is returned in
    # its entirety.
    #
    # In addition, subclasses can implement transform functions which receive a content metadata item for more
    # complicated field transformations. These functions can be content type-specific or generic for all content
    # types.
    #
    # Example:
    #     DATA_TRANSFORM_MAPPING = {
    #         'coursePrice': 'price'
    #     }
    #     # Content type-specific transformer
    #     def transform_course_price(self, course):
    #         return course['course_runs'][0]['seats']['verified]['price']
    #     # Generic transformer
    #     def transform_provider_id(self, course):
    #         return self.enterprise_configuration.provider_id
    #
    # TODO: Move this to the EnterpriseCustomerPluginConfiguration model as a JSONField.
    DATA_TRANSFORM_MAPPING = {}
    SKIP_KEY_IF_NONE = False

    def __init__(self, user, enterprise_configuration):
        """
        Initialize the exporter.
        """
        super().__init__(user, enterprise_configuration)
        self.enterprise_catalog_api = EnterpriseCatalogApiClient(self.user)

    def export(self, **kwargs):
        """
        Export transformed content metadata if there has been an update to the consumer's catalogs
        """
        enterprise_customer_catalogs = self.enterprise_configuration.customer_catalogs_to_transmit or \
            self.enterprise_customer.enterprise_customer_catalogs.all()

        content_metadata_export = {}
        force_retrieve_all_catalogs = kwargs.get('force_retrieve_all_catalogs', False)

        for enterprise_customer_catalog in enterprise_customer_catalogs:
            need_catalog_update = False
            if not force_retrieve_all_catalogs:
                # Retrieve the time at which the catalog and it's content were last updated
                content_last_modified, catalog_modified = self._get_enterprise_catalog_last_changed(
                    enterprise_customer_catalog
                )
                # Retrieve whatever modification happened more recently
                catalog_last_modified = max(
                    content_last_modified,
                    catalog_modified
                )
                last_successful_transmission = self._get_most_recent_catalog_update_time()
                # If the last successful transmission was since the last time the catalog was updated, there's no need
                # for an update
                need_catalog_update = (
                    not last_successful_transmission or (last_successful_transmission < catalog_last_modified)
                )

            if force_retrieve_all_catalogs or need_catalog_update:
                content_metadata_item = self.enterprise_catalog_api.get_content_metadata(
                    self.enterprise_customer,
                    [enterprise_customer_catalog]
                )
                for item in content_metadata_item:
                    transformed = self._transform_item(item)
                    LOGGER.debug(
                        'Exporting content metadata item with plugin configuration [%s]: [%s]',
                        self.enterprise_configuration,
                        json.dumps(transformed, indent=4),
                    )

                    # There are some scenarios where `content_last_modified` isn't present in the fetched content
                    content_last_modified = item.get('content_last_modified')
                    if content_last_modified:
                        content_last_modified = item.pop('content_last_modified')
                    else:
                        LOGGER.warning(
                            "content_last_modified field not found for {} - {} under catalog: {}".format(
                                item.get('content_type'),
                                item.get('key'),
                                enterprise_customer_catalog.uuid,
                            )
                        )

                    content_metadata_item_export = ContentMetadataItemExport(item, transformed, content_last_modified)
                    content_metadata_export[content_metadata_item_export.content_id] = content_metadata_item_export
            else:
                LOGGER.info(
                    "Skipping export of catalog:{} because no updates are needed.".format(enterprise_customer_catalog)
                )
        return OrderedDict(sorted(content_metadata_export.items()))

    def _transform_item(self, content_metadata_item):
        """
        Transform the provided content metadata item to the schema expected by the integrated channel.
        """
        content_metadata_type = content_metadata_item['content_type']
        transformed_item = {}
        for integrated_channel_schema_key, edx_data_schema_key in self.DATA_TRANSFORM_MAPPING.items():
            # Look for transformer functions defined on subclasses.
            # Favor content type-specific functions.
            transformer = (
                getattr(
                    self,
                    'transform_{content_type}_{edx_data_schema_key}'.format(
                        content_type=content_metadata_type,
                        edx_data_schema_key=edx_data_schema_key
                    ),
                    None
                )
                or
                getattr(
                    self,
                    'transform_{edx_data_schema_key}'.format(
                        edx_data_schema_key=edx_data_schema_key
                    ),
                    None
                )
            )
            if transformer:
                transformed_value = transformer(content_metadata_item)
            else:
                # The concrete subclass does not define an override for the given field,
                # so just use the data key to index the content metadata item dictionary.
                try:
                    transformed_value = content_metadata_item[edx_data_schema_key]
                except KeyError:
                    # There may be a problem with the DATA_TRANSFORM_MAPPING on
                    # the concrete subclass or the concrete subclass does not implement
                    # the appropriate field transformer function.
                    LOGGER.exception(
                        'Failed to transform content metadata item field [%s] for [%s]: [%s]',
                        edx_data_schema_key,
                        self.enterprise_customer.name,
                        content_metadata_item,
                    )
                    continue

            if transformed_value is None and self.SKIP_KEY_IF_NONE:
                continue
            transformed_item[integrated_channel_schema_key] = transformed_value

        return transformed_item

    def _get_enterprise_catalog_last_changed(self, enterprise_catalog):
        """
        Retrieves catalog metadata, specifically selecting for `content_last_modified` and `catalog_modified`.
        """
        enterprise_catalog_data = self.enterprise_catalog_api.get_enterprise_catalog(
            enterprise_catalog.uuid
        )
        content_last_modified = enterprise_catalog_data.get('content_last_modified')
        content_last_modified = dateparse.parse_datetime(
            content_last_modified
        ) if content_last_modified else datetime.min.replace(tzinfo=pytz.UTC)

        catalog_modified = enterprise_catalog_data.get('catalog_modified')
        catalog_modified = dateparse.parse_datetime(
            catalog_modified
        ) if catalog_modified else datetime.min.replace(tzinfo=pytz.UTC)
        return content_last_modified, catalog_modified

    def _get_most_recent_catalog_update_time(self):
        """
        Retrieve the last modified time of the catalog belonging to the most recent ContentMetadataItemTransmission for
        an enterprise customer.
        """
        ContentMetadataItemTransmission = apps.get_model(
            'integrated_channel',
            'ContentMetadataItemTransmission'
        )
        past_transmissions = ContentMetadataItemTransmission.objects.filter(
            enterprise_customer=self.enterprise_configuration.enterprise_customer,
            integrated_channel_code=self.enterprise_configuration.channel_code(),
            content_last_changed__isnull=False,
        ).values('content_last_changed').order_by('-content_last_changed')
        if past_transmissions:
            return past_transmissions[0]['content_last_changed']
        return None


class ContentMetadataItemExport:
    """
    Object representation of a content metadata item export.
    """

    def __init__(self, content_metadata_item, channel_content_metadata_item, content_last_changed=None):
        self.content_id = get_content_metadata_item_id(content_metadata_item)
        self.metadata = content_metadata_item
        self.channel_metadata = channel_content_metadata_item
        self.content_last_changed = content_last_changed
